;
; =============================================================================
; wndx:   multiple table (popup window) lookup utility with vertical scroll bar
;         and Ctrl-Z searches on user-selectable fields. Parameter control over
;         colors, window size, and screen placement.
; =============================================================================
;
; Original file copyright (c) 1992 Informant Communications Group and the
; author. The material here may be used in an application provided
; that this copyright/disclaimer is kept in its entirety in the source
; file. The material presented here is provided "as is" and with no guarantees.
; The Informant Communications Group/Paradox Informant and the author assume
; no responsibility for the use or misuse of the information contained herein.
;
; Contents        : procedure wndx, w1svcdx, in script "makelib.sc"
;                 : calling routine in script "tryit.sc"
; Source File     : makelib.sc, tryit.sc
; Author          : Chris Avery, CompuServe ID: [76030,375]
;
; Description     : A routine extending and enhancing a lookup table window
;                   utility published in May 1991 in the Paradox Informant by
;                   Jim Schwartz.  This routine has been generalized to handle
;                   multiple tables and window layouts, and parameterized
;                   colors, window heights, and window placement on the screen.
;                   It has been extended to include a vertical scroll indicator,
;                   and Ctrl-Z searches on user-selectable (multiple) fields.
;                   It has been improved to minimize re-writing records to the
;                   screen and to improve the response to the PgUp, PgDn, Home
;                   and End keys. The ZIP file contains a sample table, a
;                   script to build the library file "wlib", and a sample
;                   calling script to illustrate the use of this material.
;
; ============================================================================
;
; Paradox Informant                     Author: Chris Avery
; 8525 Elk Grove Blvd.                  CompuServe ID: [76030,674]
; Suite 126                             Comments (and enhancements) welcomed!
; Elk Grove, CA  95624-1777
; Phone: (916) 686-6610                 (510) 748-6001
; Fax  : (916) 686-8497                 (510) 748-6005
; BBS  : (916) 686-4740
; =========================================================================
;
; Note: procedure "wndx" remains unchanged for different windows/tables
; =========================================================================
;
PROC CLOSED wndx (wlib, wproc, ulrow, ulcol, wrows, wclr, reclr, hiclr, scrclr)
;
   PRIVATE totrecs, currec, rowcnt, currow, nxtrow, toprec, nxtrec, scrrow, scrchr, inx,
           zstr, zdone, zblink, wtxt, scrchr
;
   autolib = wlib
;
; ===>  on input:  lookup table must be "current" on entry, and it may be linked in a
;                  multi-table form. Main, CoEdit and Edit modes all OK.
; ===> on output:  returns TRUE if record selected (it will be "current" on exit),
;                  mode not changed.
;
; ===> external parameter values supplied in argument list
;
;           wlib:       library name where wproc will be found
;          wproc:       "window-proc" - custom built routine for each window, draws
;                             window using TEXT/ENDTEXT and sets internal parameters
;
;          ulrow:       upper left row to be occupied by window
;          ulcol:       upper left column to be occupied by window
;          wrows:       number of records to be displayed in window
;           wclr:       basic color attribute for window
;          reclr:       color attribute used to display records
;          hiclr:       color attribute used to display "highlighted record" and Ctrl-Z search string
;         scrclr:       color attribute used to display vertical scroll bar and indicators
;
; ===> internal paramters supplied by "wproc"
;
;         wextxt:       executable text string produces displayable "record" in variable "wtxt"
;       wfldname:       array of fieldnames in lookup table for Ctrl-Z/Alt-Z searches
;       wfldescr:       array of field descriptions
;         recrow:       row number above start of record display
;         reccol:       column number of first character in displayed record
;         scrcol:       column number of vertical scroll bar
;
; **** internal variables ****
;
   totrecs = NIMAGERECORDS()  ; records in current image of lookup table (respects limits of linked images)
   currec  = 1                ; current record in lookup table
   rowcnt  = 1                ; number of lines displayed, max is "wrows"
   currow  = 1                ; line number currently highlighted in window
   nxtrow  = 2                ; line number next to be highlighted in window
   toprec  = 1                ; recno displayed in row 1
   nxtrec  = 0                ; counter for next recno to be displayed, max is "totrecs"
   scrrow  = 1                ; current vertical scroll bar marker line
   scrchr  = CHR(223)         ; "marker" for vertical scroll bar
   zstr    = ""               ; search string for Ctrl-Z/Alt-Z
   zblink  = FALSE            ; flag to blink record on success of Ctrl-Z/Alt-Z searches
   zdone   = FALSE            ; flag is TRUE if a Ctrl-Z search just succeeded (to permit Alt-Z follow-up)
;
; draw window
;
   EXECPROC wproc             ; draw window, set parameters
;
   WHILE TRUE                 ; outside loop redisplays records after movement "outside" present display
;
      nxtrec = toprec
      MOVETO RECORD toprec
      STYLE ATTRIBUTE reclr                              ; record color
      CANVAS OFF                                         ; freeze canvas
      @recrow, 0
      WHILE (nxtrec <= totrecs AND rowcnt <= wrows)      ; display records
         EXECUTE wextxt
         ? wtxt
         nxtrec = nxtrec + 1
         rowcnt = rowcnt + 1
         DOWN
      ENDWHILE
;
      IF totrecs > wrows THEN
         STYLE ATTRIBUTE scrclr
         IF toprec > 1 THEN
            @recrow+1,reccol-2 ?? CHR(30)                ; place up arrowhead
         ELSE
            @recrow+1,reccol-2 ?? " "
         ENDIF
;
         IF rowcnt > wrows THEN
            IF toprec + wrows - 1 < totrecs THEN
               @recrow+wrows,reccol-2 ?? CHR(31)         ; place down arrowhhead
            ELSE
               @recrow+wrows,reccol-2 ?? " "
            ENDIF
         ELSE
            @recrow+wrows,reccol-2 ?? " "
            @recrow+rowcnt-1,0
         ENDIF
      ENDIF
;
      WHILE rowcnt <= wrows                              ; blank remaining lines
         STYLE ATTRIBUTE reclr
         ? SPACES (scrcol-reccol-1)
         rowcnt = rowcnt + 1
      ENDWHILE
      rowcnt = 1
;
      WHILE TRUE                                         ; inside loop traps illegal movement keys, and processes
;                                                          movemment between records currently displayed
         MOVETO RECORD currec
;
         STYLE ATTRIBUTE scrclr
         @recrow+scrrow, scrcol ?? " "                   ; clear scroll bar marker
         inx = INT (2*wrows*((currec-1)/totrecs)) + 1
         scrrow = INT ((inx-1)/2) + 1                    ; compute new marker line
         IF MOD (inx, 2) = 1 THEN
            scrchr = CHR(223)                            ; use "upper half" char
         ELSE
            scrchr = CHR(220)                            ; use "lower half" char
         ENDIF
         IF currec = totrecs THEN
            scrchr = CHR(220)                            ; correct possible rounding error for last record
         ENDIF
         @recrow+scrrow, scrcol ?? scrchr                ; place scroll bar marker
;
         CANVAS ON
;
         IF zblink THEN
            PAINTCANVAS ATTRIBUTE reclr+128              ; blink highlight on current record on z-search
               recrow+currow, reccol, recrow+currow, scrcol-2
            SLEEP 1500
            zblink = FALSE
         ENDIF
         PAINTCANVAS ATTRIBUTE hiclr                     ; stop blink
            recrow+currow, reccol, recrow+currow, scrcol-2
;
         c = GETCHAR()                                   ; get keystroke
;
         SWITCH
            CASE c = -80:                                ; Down
               zdone = FALSE
               IF currec < totrecs THEN
                  currec = currec + 1
                  nxtrow = currow + 1
                  IF nxtrow > wrows THEN
                     toprec = toprec + 1
                     currow = wrows
                     QUITLOOP
                  ENDIF
               ELSE
                  BEEP LOOP
               ENDIF
;
            CASE c = -72:                                ; Up
               zdone = FALSE
               IF currec > 1 THEN
                  currec = currec - 1
                  nxtrow = currow - 1
                  IF nxtrow < 1 THEN
                     toprec = toprec - 1
                     currow = 1
                     QUITLOOP
                  ENDIF
               ELSE
                  BEEP LOOP
               ENDIF
;
            CASE c = -73:                                ; PgUp
               zdone = FALSE
               IF currec = 1 THEN
                  BEEP LOOP
               ELSE
                  IF currow > 1 THEN
                     nxtrow = 1
                     currec = toprec
                  ELSE
                     currec = currec - wrows + 1
                     IF currec < 1 THEN
                        currec = 1
                     ENDIF
                     toprec = currec
                     QUITLOOP
                  ENDIF
               ENDIF
;
            CASE c = -81:                                ; PgDn
               zdone = FALSE
               IF currec = totrecs THEN
                  BEEP LOOP
               ELSE
                  IF currow < wrows AND currec < totrecs THEN
                     nxtrow = totrecs - toprec + 1
                     IF nxtrow > wrows THEN
                        nxtrow = wrows
                     ENDIF
                     currec = toprec + nxtrow - 1
                  ELSE
                     toprec = currec
                     currow = wrows
                     IF toprec + wrows - 1 > totrecs THEN
                        currow = totrecs - toprec + 1
                     ENDIF
                     currec = toprec + currow - 1
                     QUITLOOP
                  ENDIF
               ENDIF
;
            CASE c = -71:                                ; Home
               zdone = FALSE
               IF currec = 1 THEN
                  BEEP LOOP
               ELSE
                  currec = 1
                  IF currow > 1 AND toprec = 1 THEN
                     nxtrow = 1
                  ELSE
                     toprec = 1
                     currow = 1
                     QUITLOOP
                  ENDIF
               ENDIF
;
            CASE c = -79:                                ; End
               zdone = FALSE
               IF currec = totrecs THEN
                  BEEP LOOP
               ELSE
                  currec = totrecs
                  IF toprec + wrows > totrecs THEN
                     nxtrow = totrecs - toprec + 1
                  ELSE
                     IF totrecs >= wrows THEN
                        toprec = totrecs - wrows + 1
                        currow = wrows
                     ELSE
                        toprec = 1
                        currow = totrecs
                     ENDIF
                     QUITLOOP
                  ENDIF
               ENDIF
;
            CASE c = 13:                                 ; CR
               SETMARGIN OFF
               STYLE
               CURSOR NORMAL
               RETURN TRUE
;
            CASE c = 27:                                 ; Esc
               SETMARGIN OFF
               STYLE
               CURSOR NORMAL
               RETURN FALSE
;
            CASE c = 26:                                 ; Ctrl-Z
;
               wfld = wfldname [1]
               IF ARRAYSIZE (wfldname) > 1 THEN
                  MESSAGE "Select which field to search from list above"
                  SHOWARRAY wfldname  wfldescr
                  TO wfld
               ENDIF
               MOVETO FIELD wfld
;
               STYLE ATTRIBUTE SYSCOLOR (4)
               @0,0 ?? "Value:"
               @1,0 ?? "Enter value or pattern to search for."
               @0,8
               STYLE ATTRIBUTE reclr
               CURSOR NORMAL
               ACCEPT FIELDTYPE()
               TO zstr
               CURSOR OFF
               @0,0 CLEAR EOL
               @1,0 CLEAR EOL
;
               LOCATE PATTERN STRVAL (zstr)
               IF retval THEN
                  zblink = TRUE
                  zdone = TRUE
                  nxtrow = currow + [#] - currec
                  currec = [#]
                  IF nxtrow < 1 OR nxtrow > wrows THEN
                     toprec = currec
                     currow = 1
                     QUITLOOP
                  ENDIF
               ELSE
                  BEEP
                  MESSAGE "Match not found for \"" + zstr + "\"!"
                  SLEEP 1000
                  nxtrow = currow
               ENDIF
;
            CASE c = -44:                                ; Alt-Z
               IF zdone THEN
                  IF NOT ATLAST() THEN
                     DOWN
                     LOCATE NEXT PATTERN zstr
                     IF retval THEN
                        zblink = TRUE
                        nxtrow = currow + [#] - currec
                        currec = [#]
                        IF nxtrow < 1 OR nxtrow > wrows THEN
                           toprec = currec
                           currow = 1
                           QUITLOOP
                        ENDIF
                     ELSE
                        BEEP
                        MESSAGE "Another match not found for \"" + zstr + "\"!"
                        SLEEP 1000
                        UP
                        nxtrow = currow
                     ENDIF
                  ELSE
                     BEEP
                     MESSAGE "End of list!"
                     SLEEP 1000
                     nxtrow = currow
                  ENDIF
               ELSE
                  BEEP
                  MESSAGE "Please use Ctrl-Z for first search!"
                  SLEEP 1000
                  nxtrow = currow
               ENDIF
;
            OTHERWISE:
               BEEP LOOP
;
         ENDSWITCH
;
         CANVAS OFF
         PAINTCANVAS ATTRIBUTE reclr                     ; de-highlight current row
            recrow+currow, reccol, recrow+currow, scrcol-2
         currow = nxtrow
;
      ENDWHILE
      CANVAS OFF
   ENDWHILE
ENDPROC
;
; =======================================================================
; w1svcdx: builds window for table "SvCdx" (window design 1)
;          this proc must be hardcoded (customized) for each table/window
;          to be displayed
; =======================================================================
;
PROC w1svcdx()
;
;  ===> values supplied by DoWndx via global variables:
;
;      ulrow:      row of upper left corner of window to display
;      ulcol:      col of upper left corner of window to display
;      wrows:      number of records to display (controls height of window)
;       wclr:      color attribute for window
;     scrclr:      color attribute for scrollbar
;
;  ===> values exported via global variables:
;
;     recrow:      row number above first displayable record
;     reccol:      col number to display "record"
;     scrcol:      col number of vertical scroll bar
;     wextxt:      executable text string which produces displayable "record" for window
;   wfldname:      array of fieldnames which can be used for Ctrl-Z/Alt-Z searches
;   wfldescr:      array of field descriptions
;
;   comments:     a) total window height is wrows + 8, width is ncols (including shadows)
;                 b) hardcoded numeric values are fixed by format of window and record layout
;                         coded below.
;                 c) illegal ulrow/ulcol are corrected to place window at screen edge
;                 d) exits with canvas and cursor OFF, margin set to reccol
;
; ==> internal variables
;
   PRIVATE ncols     ; total width of table (fixed) in number of columns covered
;
; commence to begin to start:
;
   ncols = 58                                            ; table proper plus extra column plus 2 cols shadows
;
   IF ulrow < 2 THEN                                     ; trim out-of-bounds values
      ulrow = 2
   ENDIF
;
   IF ulcol < 0 THEN
      ulcol = 0
   ENDIF
;
   IF ulrow > 25 - wrows - 8 THEN                        ; 8 rows of "headers/footers/shadows" fixed by table layout
      ulrow = 25 - wrows - 8
      IF ulrow < 2 THEN
         ulrow = 2
         wrows = 15
      ENDIF
   ENDIF
;
   IF ulcol > 80 - ncols THEN
      ulcol = 80 - ncols
   ENDIF
;
   recrow  = ulrow + 2
   reccol  = ulcol + 4
   scrcol  = ulcol + ncols - 7
;
;
   ARRAY wfldname [3]
   ARRAY wfldescr [3]
;
   wfldname [1] = "SvCd"
   wfldescr [1] = "Search treatment code field"
   wfldname [2] = "Descr"
   wfldescr [2] = "Search description field"
   wfldname [3] = "Amt"
   wfldescr [3] = "Search treatment charge field"
;
   wextxt = "wtxt =  SUBSTR (\" \" + [SvCd]  + SPACES (10), 1, 11) + " +
                   " SUBSTR ([Descr] + SPACES (27), 1, 27) + " +
                   " FORMAT (\"W7.2, AR\", [Amt]) + \" \""
;
   CURSOR OFF
   CANVAS OFF
;
   STYLE ATTRIBUTE wclr
   SETMARGIN ulcol
   @ulrow,0
;
   TEXT
  Service Codes ͻ
   Code       Description                Charge     
 Ķ
   ENDTEXT
;
   @recrow, 0
   FOR i FROM 1 TO wrows
       ? "                                                    "
   ENDFOR
;
; format= ___CCCCCCCCC__DDDDDDDDDDDDDDDDDDDDDDDDD__AAAAAAAA_V_
;
   @recrow+wrows+1, 0
;
   TEXT
 Ķ
  Press "Enter" to select, or "Esc" to quit.        
  Press Ctrl-Z to search for a particular value     
 ͼ
   ENDTEXT
   STYLE
;
   PAINTCANVAS FILL " " ATTRIBUTE scrclr
      recrow+1, scrcol, recrow+wrows, scrcol                   ; vertical ruler
   PAINTCANVAS FILL " " ATTRIBUTE wclr
      ulrow, scrcol+3, ulrow+wrows+6, scrcol+3                 ; right vertical table extension
   PAINTCANVAS FILL chr(176) ATTRIBUTE wclr
      ulrow+1, scrcol+4, ulrow+wrows+6, scrcol+5               ; right vertical shadow (2 cols)
   PAINTCANVAS FILL chr(176) ATTRIBUTE wclr
      ulrow+wrows+7, ulcol+2, ulrow+wrows+7, scrcol+5          ; lower border shadow (1 char)
;
   SETMARGIN reccol
;
ENDPROC
;
;
CREATELIB "wlib"
WRITELIB  "wlib"  wndx, w1svcdx
